/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/mm.h>
#include <asm/io.h>     /* for virt_to_phys */
#include "pvrusb2.h"
#include "pvrusb2-buffer.h"

/* 
 * pvr_buff_complete()
 *
 * This is called after an urb used for video stream is back.
 *
 */

static void pvr_buffer_complete (struct urb *purb, struct pt_regs *regs)
{
	struct pvr_buffer *b = (struct pvr_buffer *) purb->context;
	struct pvr *pvr = b->pvr;
	void *buff = purb->transfer_buffer;
    unsigned long flags;

//    pvr_debug("pvr_buffer_complete");
    
	if (purb->status != -ENOENT) {
        /* urb was not canceled. Put it in our rx ring. */
        b->data = buff;
        b->size = purb->actual_length;
        spin_lock_irqsave(&pvr->ring_lock, flags);
        pvr->rx_buffer[pvr->rx2 % PVR_NBUFF] = b;
        pvr->rx2++;
        spin_unlock_irqrestore(&pvr->ring_lock, flags);
        if (waitqueue_active(&pvr->wait_data))
		    wake_up (&pvr->wait_data);
    }
    if (atomic_dec_and_test(&pvr->submitted_urbs)){
        pvr_debug("pvr_buffer_complete : no more urbs submitted !");
        wake_up(&pvr->urbs_collected);
    }
    b->submitted = 0;
}

int pvr_alloc_buffer (struct pvr *pvr)
{
    int i, pipe;
    struct pvr_buffer *b;

    pvr_debug("pvr_alloc_buffer");

    if (pvr->mem_state != _freed){
        err("illegal pvr_alloc_buffer");
        return -EBUSY;
    }

    memset(pvr->buffer, 0, sizeof(struct pvr_buffer)*PVR_NBUFF);

    if (pvr->buffer_count > PVR_NBUFF){
        err("to many video buffers!");
        return -ENOMEM;
    }

    pvr->mem_state = _allocated;

    for (i = 0; i < pvr->buffer_count; i++){
        b = &pvr->buffer[i];
        b->purb = usb_alloc_urb(0, GFP_KERNEL);
        if (b->purb == NULL)
            return -ENOMEM;
        /* remember, we want page aligned buffers for mmap */
        b->buff = kmalloc(PVR_VID_BUFFSIZE + PAGE_SIZE - 1, GFP_KERNEL);
        if (b->buff == NULL)
            return -ENOMEM;
        b->pvr = pvr;
        pipe = usb_rcvbulkpipe(pvr->udev, PVR_VID_ENDPOINT);
        usb_fill_bulk_urb (b->purb, pvr->udev, pipe,
                           (void *) PAGE_ALIGN((unsigned long)b->buff),
                           PVR_VID_BUFFSIZE, pvr_buffer_complete, b);
      	b->purb->transfer_flags = 0;
        b->submitted = 0;
        b->map = NULL;
    }

    return 0;
}

/* Safe even if nothing allocated */

void pvr_free_buffer(struct pvr *pvr)
{
    int i;
    struct pvr_buffer *b;

    pvr_debug("pvr_free_buffer");
    for (i = 0; i < PVR_NBUFF; i++){
        b = &pvr->buffer[i];
        if (b->purb){
            usb_free_urb(b->purb);
            b->purb = NULL;
            if (b->buff)
                kfree(b->buff);
        }
    }
    pvr->mem_state = _freed;
}

/* Call with lock held */

void pvr_unlink_all_buffer(struct pvr *pvr)
{
    int i;

    pvr_debug("pvr_unlink_all_buffer");

    for(i = 0 ; i < pvr->buffer_count; i++)
        if (pvr->buffer[i].purb)
            usb_unlink_urb(pvr->buffer[i].purb);

    /* Is there a small chance that an urb is back, but his completion
     * handler not yet complete ? 
     * We must wait a bit to be sure we have everybody...
     */

    if (wait_event_interruptible_timeout(pvr->urbs_collected,
                                         atomic_read(&pvr->submitted_urbs) == 0,
                                         HZ) == 0)
        err("timeout waiting for urbs !");

    /* reinitialize the rx ring. */
    pvr->rx1 = 0;
    pvr->rx2 = 0;
}

/* Call with lock held */

void pvr_submit_all_buffer(struct pvr *pvr)
{
    int i;

    pvr_debug("pvr_submit_all_buffer");

    if (pvr->disconnecting)
        return;

    atomic_set(&pvr->submitted_urbs,pvr->buffer_count);
    for (i = 0 ; i < pvr->buffer_count ; i++) {
        pvr->buffer[i].submitted = 1;
        usb_submit_urb(pvr->buffer[i].purb, GFP_KERNEL);
    }
}

int pvr_submit_buffer(struct pvr *pvr, int i)
{
//    pvr_debug("pvr_submit_buffer");

    if (pvr->buffer[i].submitted || pvr->disconnecting)
        return -EINVAL;
    
    atomic_inc(&pvr->submitted_urbs);
    pvr->buffer[i].submitted = 1;
    return usb_submit_urb(pvr->buffer[i].purb, GFP_KERNEL);
}

static void pvr_vm_open (struct vm_area_struct *vma)
{
	struct pvr_mapping *map = vma->vm_private_data;

    pvr_debug("pvr_vm_open");

	map->count++;
}

static void pvr_vm_close(struct vm_area_struct *vma)
{
    int i;
	struct pvr_mapping *map = vma->vm_private_data;

    pvr_debug("pvr_vm_close");

    map->count--;

    if (map->count == 0) {
        struct pvr *pvr = map->pvr;

        pvr_debug("pvr_vm_close : umapping memory");

        for (i = 0 ; i < pvr->buffer_count ; i++) {
            if (pvr->buffer[i].map == map)
                pvr->buffer[i].map = NULL;
        
        }
        kfree(map);
        down(&pvr->sem);
        pvr->map_count--;
        if (pvr->map_count == 0)
            pvr->mem_state = _allocated;
        up(&pvr->sem);
    }
}


static struct vm_operations_struct pvr_vm_ops = {
	.open = pvr_vm_open,
	.close = pvr_vm_close,
};

/* Call with lock held. For V4L1 compat, we need to be able to map all
 * buffers in one shot.
 */

int pvr_map_buffer(struct pvr *pvr, struct vm_area_struct *vma)
{
    int i, first, last;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long size = vma->vm_end - vma->vm_start;
    unsigned long start = vma->vm_start;
    void *buff;
    struct pvr_mapping *map;

    pvr_debug("pvr_map_buffer");

    if (pvr->mem_state < _allocated) {
        err("illegal pvr_map_buffer");
        return -EINVAL;
    }

    if (size == 0 || size % PVR_VID_BUFFSIZE != 0){
        err("pvr_map_buffer : wrong size !");
        return -EINVAL;
    }

    /* find first and last buffer to remap */
    first = offset / PVR_VID_BUFFSIZE;
    last = first + size / PVR_VID_BUFFSIZE - 1;

    /* Verify none of them is already mapped. */
    for (i = first ; i <= last ; i++)
        if (pvr->buffer[i].map){
            err("buffer %i already mapped", i);
            return -EBUSY;
        }
    /* map'em all */
    map = kmalloc(sizeof(struct pvr_mapping), GFP_KERNEL);
    if (map == NULL)
        return -ENOMEM;
    map->pvr = pvr;
    map->count = 1;
    vma->vm_private_data = map;
    vma->vm_ops = &pvr_vm_ops;
    vma->vm_flags |= VM_RESERVED | VM_SHARED;
    pvr->map_count++;
    pvr->mem_state = _mapped;

    for (i = first ; i <= last ; i++) {
        buff = (void *) PAGE_ALIGN((unsigned long)pvr->buffer[i].buff);
        if (remap_page_range(vma, start, virt_to_phys(buff),
                             PVR_VID_BUFFSIZE, PAGE_SHARED))
            return -EAGAIN;
        pvr->buffer[i].map = map;
        start += PVR_VID_BUFFSIZE;
    }

    return 0;
}

/*
 * Wait for a buffer to be available. Timeout of 0 means non blocking io.
 * Returns buffer index on success,  < 0 error code on failure.
 * Detects stream break.
 * Lock outside. 
 */

int pvr_buffer_wait(struct pvr *pvr, unsigned long timeout)
{
    
    if (timeout) {
        wait_event_interruptible_timeout(pvr->wait_data,
                                         pvr->rx2 - pvr->rx1 != 0,
                                         timeout);
        if (pvr->rx2 - pvr->rx1 != 0) {
            pvr->last_timestamp = jiffies;
            return pvr->rx1 % PVR_NBUFF;
        } else
            return -EIO;
    } else {
        if (pvr->rx2 - pvr->rx1){
            pvr->last_timestamp = jiffies;
            return pvr->rx1 % PVR_NBUFF;
        } else {
            if (time_after(jiffies, pvr->last_timestamp + 3 * HZ))
                /* broken stream. Now it's time to kiss your ass goodbye... */
                return -EIO;
            else
                return -EAGAIN;
        }
    }
}

int pvr_cgmbuf(struct pvr *pvr, struct video_mbuf *vmb)
{
    int i, off;

    pvr_debug("pvr_cgmbuf");

    down(&pvr->sem);

    pvr_unlink_all_buffer(pvr);
    if (pvr->buffer_count > VIDEO_MAX_FRAME) {
        if (pvr->mem_state > _freed)
            pvr_free_buffer(pvr);
        pvr->buffer_count = VIDEO_MAX_FRAME;
        if (pvr_alloc_buffer(pvr)){
            up(&pvr->sem);
            return -ENOMEM;
        }
    }
    pvr->io_method = PVR_IO_MMAP; 

    up(&pvr->sem);

    vmb->size = pvr->buffer_count * PVR_VID_BUFFSIZE;
    vmb->frames = pvr->buffer_count;
    for (i = 0, off = 0 ; i < pvr->buffer_count ; i++, off += PVR_VID_BUFFSIZE)
        vmb->offsets[i] = off;

pvr_debug("pvr_cgmbuf done");
    return 0;
}


int pvr_reqbufs(struct pvr *pvr, struct v4l2_requestbuffers *vr)
{
    int ret = 0;

    pvr_debug("pvr_reqbufs");

    if (vr->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;
    
    if (vr->memory != V4L2_MEMORY_MMAP)
        return -EINVAL;

    if (vr->count > PVR_NBUFF)
        vr->count = PVR_NBUFF;

    down(&pvr->sem);

    if (vr->count) {
        /* alloc or re-alloc */
        switch(pvr->mem_state){
            case _mapped:
                 ret = -EBUSY;
                 break;
            case _allocated:
                pvr_unlink_all_buffer(pvr);
                pvr_free_buffer(pvr);
            case _freed:
                pvr->buffer_count = vr->count;
                if (pvr_alloc_buffer(pvr))
                    vr->count = 0;
                else
                    pvr->io_method = PVR_IO_MMAP;
        }
    } else {
        if (pvr->mem_state == _mapped)
            ret = -EBUSY;
        else if (pvr->mem_state == _allocated)
            /* Make sure we are not streaming, then free all buffers */
            pvr_state_below(pvr, _stopped);
            pvr_unlink_all_buffer(pvr);
            pvr_free_buffer(pvr);
            pvr->io_method = PVR_IO_READ;
    }

    up(&pvr->sem);

    return ret;
}

int pvr_querybuf(struct pvr *pvr, struct v4l2_buffer *vb)
{
    if (vb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;

    if (vb->index < 0 || vb->index >= pvr->buffer_count)
        return -EINVAL;

    vb->memory = V4L2_MEMORY_MMAP;
    vb->flags = 0;

    down(&pvr->sem);

    if (pvr->buffer[vb->index].map)
        vb->flags |= V4L2_BUF_FLAG_MAPPED;

    if (pvr->buffer[vb->index].submitted)
        vb->flags |= V4L2_BUF_FLAG_QUEUED;
    else
        vb->flags |= V4L2_BUF_FLAG_DONE;

    vb->length = PVR_VID_BUFFSIZE;
    vb->m.offset = vb->index * PVR_VID_BUFFSIZE;

    up(&pvr->sem);

    return 0;
}

int pvr_qbuf(struct pvr *pvr, struct v4l2_buffer *vb)
{
    int ret= 0;

    if (vb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ||
        vb->memory != V4L2_MEMORY_MMAP)
        return -EINVAL;

    if (vb->index < 0 || vb->index >= pvr->buffer_count)
        return -EINVAL;

    down(&pvr->sem);
    ret = pvr_submit_buffer(pvr, vb->index);
    up(&pvr->sem);

    if (ret == 0){
        vb->flags |= V4L2_BUF_FLAG_MAPPED | V4L2_BUF_FLAG_QUEUED;
        vb->flags &= ~V4L2_BUF_FLAG_DONE;
    }

    return ret ? -EINVAL : 0;
}

int pvr_dqbuf(struct pvr *pvr, struct v4l2_buffer *vb, int nonblock)
{
    int ret;

    if (vb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;

//     if (vb->memory != V4L2_MEMORY_MMAP)
//        return -EINVAL;

    if (vb->index < 0 || vb->index >= pvr->buffer_count)
        return -EINVAL;

    down(&pvr->sem);

    ret = pvr_buffer_wait(pvr, nonblock ? 0 : 2 * HZ);
    if (ret >= 0)
        pvr->rx1++;

    up(&pvr->sem);

    return (ret < 0) ? ret : 0;
}

int pvr_streamon(struct pvr *pvr)
{
    int ret = 0;

    pvr_debug("pvr_streamon");

    if (pvr->mem_state < _allocated || atomic_read(&pvr->submitted_urbs) == 0)
        return -EINVAL;

    down(&pvr->sem);

    pvr_state_above(pvr, _streaming);
    /* wait 2 seconds for data. */
    if (pvr_buffer_wait(pvr, 2 * HZ) < 0){
        /* timeout. retry. */
        pvr_debug("pvr_streamon : retrying");
        pvr_state_below(pvr, _reset);
        pvr_state_above(pvr, _streaming);
        if (pvr_buffer_wait(pvr, 2 * HZ) < 0)
            ret = -EINVAL;
    }

    up(&pvr->sem); 

    return ret;
}

int pvr_streamoff(struct pvr *pvr)
{
    pvr_debug("pvr_streamoff");

    down(&pvr->sem);

    pvr_state_below(pvr, _stopped);

    pvr_unlink_all_buffer(pvr);

    /* Reset received buffer ring. */
    pvr->rx1 = 0;
    pvr->rx2 = 0;

    up(&pvr->sem);
    
    return 0;
}
