/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/device.h>   // for linux/firmware.h
#include <linux/firmware.h>
#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"

/* 
 * pvr_upload_firmware2()
 *
 * This uploads encoder firmware on endpoint 2.
 *
 */

int pvr_upload_firmware2(struct pvr *pvr)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe, fw_len, fw_done;
    int actual_length;
    int ret = 0;

    /* First prepare firmware loading */
    ret |= pvr_write_u8(pvr, 0x0b, 1);
    ret |= pvr_write_u16(pvr, 0xde00, 0);
    ret |= pvr_write_u8(pvr, 0xeb, 1);
    ret |= pvr_write_register(pvr, 0x0048, 0xffffffff); /* interrupt mask */
    ret |= pvr_write_register(pvr, 0x9020, 0x88000000); /* gpio dir */
    ret |= pvr_write_register(pvr, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr_write_u8(pvr, 0xdd, 0);
    ret |= pvr_write_register(pvr, 0xa064, 0x00000000); /* APU command */
    ret |= pvr_write_register(pvr, 0x9020, 0x08040000); /* gpio dir */
    ret |= pvr_write_register(pvr, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr_write_register(pvr, 0x9058, 0xedffffff); /* VPU ctrl */
    ret |= pvr_write_register(pvr, 0x9054, 0xfdffffff); /* reset hw blocks */
    ret |= pvr_write_register(pvr, 0x07f8, 0x00080080); /* encoder SDRAM refresh */
    ret |= pvr_write_register(pvr, 0x07fc, 0x1a000000); /* encoder SDRAM pre-charge */
    ret |= pvr_write_register(pvr, 0x0700, 0x00000000); /* I2C clock */
    ret |= pvr_write_register(pvr, 0xaa00, 0x00000000); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa04, 0x10780500); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa10, 0x00851400); /* unknown */
    ret |= pvr_write_register(pvr, 0xaa18, 0x00008400); /* unknown */
    ret |= pvr_write_u8(pvr, 0x52, 0);

    if (ret)
        return ret;

    /* Now send firmware */
    
    ret = request_firmware(&fw_entry, FIRMWARE2_FILE, &pvr->udev->dev);

    if (ret) {
        err("request_firmware failed for '%s'", FIRMWARE2_FILE);
        return ret;
    }

    fw_len = fw_entry->size;

    if (fw_len % FIRMWARE_CHUNK_SIZE) {
        err("size of %s must be a multiple of 8192B", FIRMWARE2_FILE);
        release_firmware(fw_entry);
        return -1;
    }

    fw_ptr = kmalloc(FIRMWARE_CHUNK_SIZE, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    pipe = usb_sndbulkpipe(pvr->udev, PVR_FIRMWARE_ENDPOINT);

    for (fw_done = 0 ; (fw_done < fw_len) && !ret ; fw_done += FIRMWARE_CHUNK_SIZE ) {
        int i;
        memcpy(fw_ptr, fw_entry->data + fw_done, FIRMWARE_CHUNK_SIZE);
        /* Usbsnoop log  shows that we must swap bytes... */
        for (i = 0; i < FIRMWARE_CHUNK_SIZE/4 ; i++)
           ((u32 *)fw_ptr)[i] = ___swab32(((u32 *)fw_ptr)[i]);

        ret |= usb_bulk_msg(pvr->udev, pipe, fw_ptr, FIRMWARE_CHUNK_SIZE,
                            &actual_length, HZ);
        ret |= (actual_length != FIRMWARE_CHUNK_SIZE);
    }

    pvr_debug("upload of %s : %i / %i ", FIRMWARE2_FILE, fw_done, 
              fw_entry->size);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    if (ret)
        return ret;

    /* Finish upload */

    ret |= pvr_write_register(pvr, 0x9054, 0xffffffff); /* reset hw blocks */
    ret |= pvr_write_register(pvr, 0x9058, 0xe8ffffff); /* VPU ctrl */
    ret |= pvr_write_u16(pvr, 0x0600, 0);

    return ret;
}

struct pvr_cmd post_encoder_cmd [] = {
    PVR_CMD(0,    "\x01\x03\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x02\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x54"),
    PVR_CMD(0,    "\x01\x00\x00\x00\x00\x00\x00\x44"),
    PVR_CMD(0, "")
};

#define PVR_DECOMPOSE_LE(t,i,d)   (t)[i] = (d) & 0xff;\
                                  (t)[i+1] = ((d) >> 8) & 0xff;\
                                  (t)[i+2] = ((d) >> 16) & 0xff;\
                                  (t)[i+3] = ((d) >> 24) & 0xff

static int pvr_write_encoder_cmd(struct pvr *pvr, u8 cmd, u32 data1,
                              u32 data2, u32 data3, u32 data4)
{
    int ret = 0;

    pvr_debug("pvr_write_encoder_cmd");

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    pvr->cmd_buffer[0] = 0x01;
    pvr->cmd_buffer[7] = 0x44;
    pvr->cmd_buffer[8] = cmd;
    pvr->cmd_buffer[14] = 0x45;
    pvr->cmd_buffer[21] = 0x46;
    pvr->cmd_buffer[28] = 0x47;
    PVR_DECOMPOSE_LE(pvr->cmd_buffer,29,data1);
    pvr->cmd_buffer[35] = 0x48;
    PVR_DECOMPOSE_LE(pvr->cmd_buffer,36,data2);
    pvr->cmd_buffer[42] = 0x49;
    PVR_DECOMPOSE_LE(pvr->cmd_buffer,43,data3);
    pvr->cmd_buffer[49] = 0x4a;
    PVR_DECOMPOSE_LE(pvr->cmd_buffer,50,data4);
    pvr->cmd_buffer[56] = 0x4b;

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 57, NULL, 0);

    ret |= pvr_send_sequence(pvr, post_encoder_cmd);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_configure_encoder(struct pvr *pvr)
{
    int ret = 0, audio;

    /* set stream output port. */
    ret |= pvr_write_encoder_cmd(pvr, 0xbb, 0x01, 0x01, 0, 0);

    /* set the Program Index Information. We want I,P,B frames (max 400) */
    ret |= pvr_write_encoder_cmd(pvr, 0xc7, 0x07, 0x0190, 0, 0);

    /* NOTE : windows driver sends some 0xdc */

    /* Strange compared to ivtv data. */
    ret |= pvr_write_encoder_cmd(pvr, 0xd6, 0x0120, 0x0120, 0, 0);

    /* setup firmware to notify us about some events (don't know why...) */
    ret |= pvr_write_encoder_cmd(pvr, 0xd5, 0, 0, 0x10000000, 0xffffffff);

    /* set fps to 25 or 30 (1 or 0)*/
    ret |= pvr_write_encoder_cmd(pvr, 0x8f, (pvr->standard != PVR_STD_NTSC),
                                0, 0, 0);
    /* set encoding resolution */
    ret |= pvr_write_encoder_cmd(pvr, 0x91, pvr->video_height,
                                pvr->video_width, 0, 0);
    /* set encoding aspect ratio to 4:3 */
    ret |= pvr_write_encoder_cmd(pvr, 0x99, 0x02, 0, 0, 0);

    /* disable VBI line information */
    ret |= pvr_write_encoder_cmd(pvr, 0xb7, 0xffffffff, 0, 0, 0);

    /* set stream type, depending on resolution. */
    ret |= pvr_write_encoder_cmd(pvr, 0xb9, (pvr->video_width == PVR_WIDTH_VCD) ?
                               0x0b : 0x0a, 0, 0, 0);
    /* set video bitrate */
    ret |= pvr_write_encoder_cmd(pvr, 0x95, pvr_tbl_vbr[pvr->video_vbr],
                                 pvr->video_bitrate,
                                 (int) pvr->video_peakbitrate / 400, 0);
    /* setup GOP structure (GOP size = 0f or 0c, 3-1 = 2 B-frames) */
    ret |= pvr_write_encoder_cmd(pvr, 0x97, (pvr->standard == PVR_STD_NTSC) ?
                                0x0f : 0x0c, 0x03, 0, 0);
    /* enable 3:2 pulldown */
    ret |= pvr_write_encoder_cmd(pvr, 0xb1, 0, 0, 0, 0);

    /* set GOP open/close property (open) */
    ret |= pvr_write_encoder_cmd(pvr, 0xc5, 0, 0, 0, 0);

    /* set audio stream properties */
    audio =   pvr_tbl_audiobitrate[pvr->audio_bitrate]
            | pvr_tbl_srate[pvr->audio_srate]
            | pvr->audio_layer << 2 
            | pvr_tbl_crc[pvr->audio_crc]
            | pvr_tbl_emphasis[pvr->audio_emphasis];

    pvr_debug("audio : %02x", audio);
    ret |= pvr_write_encoder_cmd(pvr, 0xbd, audio, 0, 0, 0);

    /* set dynamic noise reduction filter to manual, Horiz/Vert */
    ret |= pvr_write_encoder_cmd(pvr, 0x9b, 0, 0x03, 0, 0);

    /* dynamic noise reduction filter param */
    ret |= pvr_write_encoder_cmd(pvr, 0x9d, 0, 0, 0, 0);

    /* dynamic noise reduction median filter */
    ret |= pvr_write_encoder_cmd(pvr, 0x9f, 0, 0xff, 0, 0xff);

    /* spacial prefiler parameter */
    ret |= pvr_write_encoder_cmd(pvr, 0xa1, 0x01, 0x01, 0, 0);
    
    /* initialize video input */
    ret |= pvr_write_encoder_cmd(pvr, 0xcd, 0, 0, 0, 0);

    return ret;
}

int pvr_encoder_start(struct pvr* pvr)
{
    return pvr_write_encoder_cmd(pvr, 0x81, 0, 0x13, 0, 0);
//     return pvr_write_encoder_cmd(pvr, 0x81, 0x01, 0x17, 0, 0);
}

int pvr_encoder_stop(struct pvr* pvr)
{
    return pvr_write_encoder_cmd(pvr, 0x82, 0x01, 0, 0x13, 0);
//     return pvr_write_encoder_cmd(pvr, 0x82, 0x01, 0x01, 0x17, 0);
}
