/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/signal.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

#include <linux/videodev.h>
#include <linux/firmware.h>

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-remote.h"
#include "pvrusb2-saa7115.h"
#include "pvrusb2-encoder.h"
#include "pvrusb2-msp34x5g.h"
#include "pvrusb2-tuner.h"
#include "pvrusb2-eeprom.h"
#include "pvrusb2-buffer.h"

#define DRIVER_VERSION "0.6(08/09/2004)"
#define DRIVER_AUTHOR "Aurelien Alleaume <slts@free.fr>"
#define DRIVER_DESC "Hauppauge WinTV-PVR-USB2 MPEG2 Encoder/Tuner"

static int video_nr = -1;       // Number of Video Devices Initialized
static int pvr_nr = 0;          // sequential number of pvrusb2 device

module_param(video_nr, int, -1);
MODULE_PARM_DESC(video_nr, "Offset for device's minor");

static struct usb_device_id pvr_table[] = {
	{ USB_DEVICE(0x2040, 0x2900) },
	{ }
};

/* 
 * pvr_reset_ctl_endpoints().
 *
 * Reset bulk endpoints used to transmit commands.
 *
 */

static void pvr_reset_ctl_endpoints(struct pvr *pvr)
{
    usb_settoggle(pvr->udev, PVR_CTL_WRITE_ENDPOINT & 0xf,
                  !(PVR_CTL_WRITE_ENDPOINT & USB_DIR_IN), 0);
    usb_settoggle(pvr->udev, PVR_CTL_READ_ENDPOINT & 0xf,
                  !(PVR_CTL_READ_ENDPOINT & USB_DIR_IN), 0);
    usb_clear_halt(pvr->udev, usb_rcvbulkpipe(pvr->udev,
                                             PVR_CTL_READ_ENDPOINT & 0x7f));
    usb_clear_halt(pvr->udev, usb_sndbulkpipe(pvr->udev,
                                             PVR_CTL_WRITE_ENDPOINT & 0x7f));
}

/* 
 * pvr_send_request().
 *
 * Can be used to send or receive data or both. ctl_sem must be held.
 *
 */

int pvr_send_request(struct pvr *pvr, void *data,
                            unsigned int dlen, void *res,
                            unsigned int rlen)
{
    int ret = 0, actual_dlen;
    unsigned int pipe;

//    pvr_debug("pvr_send_request : %i", dlen);

    if (pvr->disconnecting || pvr->dev_state == _unstable)
        return -EIO;

    /* send the data if any. */
    if(dlen && data){
        pipe = usb_sndbulkpipe(pvr->udev, PVR_CTL_WRITE_ENDPOINT);
        ret = usb_bulk_msg(pvr->udev, pipe, data, dlen, &actual_dlen, HZ / 4);
        if (ret){
            info("transaction failed, retrying"); 
            pvr_reset_ctl_endpoints(pvr);
            /* retry */
            ret = usb_bulk_msg(pvr->udev, pipe, data, dlen, &actual_dlen, HZ / 4);
            if (ret)
                pvr_state_change(pvr, _unstable);
        }
    }

    if(rlen == 0)
        return 0;

    pipe = usb_rcvbulkpipe(pvr->udev, PVR_CTL_READ_ENDPOINT);
    ret = usb_bulk_msg(pvr->udev, pipe, res, rlen, &actual_dlen, HZ / 4);

    if (ret){
        info("transaction failed, retrying"); 
        pvr_reset_ctl_endpoints(pvr);
        /* retry */
        ret = usb_bulk_msg(pvr->udev, pipe, res, rlen, &actual_dlen, HZ / 4);
        if (ret)  
            pvr_state_change(pvr, _unstable);
    }

    return ret;
}

/* 
 * pvr_send_sequence().
 *
 * executes a sequence of commands ended by a zero length one.
 *
 */

int pvr_send_sequence(struct pvr *pvr, struct pvr_cmd *cmd)
{
    int ret = 0, i;

//    pvr_debug("pvr_send_sequence");

    i = 0;
    while ((cmd[i].length != 0) && (ret == 0)){
        memcpy(pvr->cmd_buffer, cmd[i].cmd, cmd[i].length);
        ret |= pvr_send_request(pvr, pvr->cmd_buffer, cmd[i].length,
                                pvr->cmd_buffer, cmd[i].ret);
        i++;
    }
    return ret;
}

/* 
 * pvr_upload_firmware1().
 *
 * It will upload the firmware to the inboard USB controller (FX2) ; Making
 * it re-enumerate.
 *
 * NOTE : the pointer to the firmware data given by request_firmware()
 * is not suitable for an usb transaction.
 *
 */

static int pvr_upload_firmware1(struct usb_device *udev, char *firmware)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe;
    int ret;
    u16 address;

    pvr_debug("pvr_upload_firmware1");

    usb_settoggle(udev, 0 & 0xf, !(0 & USB_DIR_IN), 0);
    usb_clear_halt(udev, usb_sndbulkpipe(udev, 0 & 0x7f));

    pipe = usb_sndctrlpipe(udev, 0);
    ret = request_firmware(&fw_entry, firmware, &udev->dev);
    
    if (ret) {
        err("request_firmware failed for '%s'", firmware);
        return ret;
    }

    if (fw_entry->size != 0x2000){
        err("wrong fx2 firmware size");
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    fw_ptr = kmalloc(0x800, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    /*
     * Write 0x01 to the CPUCS register at address e600 to stop the CPU 
     * during firmware upload.
     */
    
    ((u8*) fw_ptr)[0] = 0x01;
    ret = usb_control_msg(udev, pipe, 0xa0, 0x40, 0xe600, 0, fw_ptr, 1, HZ);

    /*
     * upload the firmware to address 0000-1fff in 2048 (=0x800) bytes chunk.
     */

    for(address = 0; address < fw_entry->size; address += 0x800){ 
        memcpy(fw_ptr, fw_entry->data + address, 0x800);
        ret += usb_control_msg(udev, pipe, 0xa0, 0x40, address,
                               0, fw_ptr, 0x800, HZ);
    }

    /*
     * Release the CPU from reset state by writing 0x00 to CPUCS.
     */
    ((char*) fw_ptr)[0] = 0x00;
    ret += usb_control_msg(udev, pipe, 0xa0, 0x40, 0xe600, 0, fw_ptr, 1, HZ);
    pvr_debug("CPUCS register restart command : %d byte written", ret);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    /* We should have written 1 + 8192 + 1 bytes */
    return (ret == 8194) ? 0 : -EIO;
}

int pvr_write_register(struct pvr *pvr, u16 reg, u32 data)
{
    int ret;

    pvr_debug("pvr_write_register");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = 0x04;  /* write register prefix */
    PVR_DECOMPOSE_BE(pvr->cmd_buffer,1,data);
    pvr->cmd_buffer[5] = 0;
    pvr->cmd_buffer[6] = (reg >> 8) & 0xff;
    pvr->cmd_buffer[7] = reg & 0xff;


    ret = pvr_send_request(pvr, pvr->cmd_buffer, 8, pvr->cmd_buffer, 0);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_read_register(struct pvr *pvr, u16 reg, u32 *data)
{
    int ret = 0;

    pvr_debug("pvr_read_register");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = 0x05;  /* read register prefix */
    pvr->cmd_buffer[1] = 0;
    pvr->cmd_buffer[2] = 0;
    pvr->cmd_buffer[3] = 0;
    pvr->cmd_buffer[4] = 0;
    pvr->cmd_buffer[5] = 0;
    pvr->cmd_buffer[6] = (reg >> 8) & 0xff;
    pvr->cmd_buffer[7] = reg & 0xff;

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 8, pvr->cmd_buffer, 4);
    *data = *((u32 *)pvr->cmd_buffer) & 0xffff;

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_write_u16(struct pvr *pvr, u16 data, int res)
{
    int ret;

    pvr_debug("pvr_write_u16");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = (data >> 8) & 0xff;
    pvr->cmd_buffer[1] = data & 0xff;

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 2, pvr->cmd_buffer, res);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_write_u8(struct pvr *pvr, u8 data, int res)
{
    int ret;

    pvr_debug("pvr_write_u8");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = data;

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 1, pvr->cmd_buffer, res);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_i2c_write(struct pvr *pvr, u8 i2c_addr, u8 *data, u8 length)
{
    int ret = 0;

//    pvr_debug("pvr_i2c_write");

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);

    pvr->cmd_buffer[0] = 0x08;  /* write prefix */
    pvr->cmd_buffer[1] = i2c_addr;  /* i2c addr of chip */
    pvr->cmd_buffer[2] = length;  /* length of what follows */

    memcpy(pvr->cmd_buffer + 3, data, length);

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, length + 3, pvr->cmd_buffer, 1);
    /* FIXME : we might want to check pvr->cmd_buffer[0] for write succeed  (08 ?) */
    up(&pvr->ctl_sem);

    return ret;
}

int pvr_i2c_read(struct pvr *pvr, u8 i2c_addr, u8 *data, u8 dlen, u8 *res, u8 rlen)
{
    int ret;

//    pvr_debug("pvr_i2c_read");

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    
    pvr->cmd_buffer[0] = 0x09;  /* read prefix */
    pvr->cmd_buffer[1] = dlen;  /* arg length */
    pvr->cmd_buffer[2] = rlen;  /* anwer length. Device will send one more byte (status). */
    pvr->cmd_buffer[3] = i2c_addr;  /* i2c addr of chip */

    if (data && dlen)
        memcpy(pvr->cmd_buffer + 4, data, dlen);

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 4 + dlen, pvr->cmd_buffer, rlen + 1);

    if (res)
        memcpy(res, pvr->cmd_buffer + 1, rlen);

    up(&pvr->ctl_sem);

    return ret;
}

/*
 * pvr_do_command()
 *
 * Executes a simple command.
 */

int pvr_do_command(struct pvr *pvr, struct pvr_cmd *cmd)
{
    int ret;

    pvr_debug("pvr_do_command");

    if (cmd == NULL) 
        return -EIO;

    down(&pvr->ctl_sem);

    /* data command (cmd->cmd) is (often) statically allocated. For usb
     *  transfer we need dynamic memory.
    */
    memcpy(pvr->cmd_buffer, cmd->cmd, cmd->length);

    ret = pvr_send_request(pvr, pvr->cmd_buffer, cmd->length,
                           pvr->cmd_buffer, cmd->ret);

    up(&pvr->ctl_sem);

    return ret;
}

static int pvr_configure_capture(struct pvr *pvr)
{
    int ret = 0;

    ret |= pvr_configure_decoder(pvr);

    ret |= pvr_configure_encoder(pvr);

    ret |= pvr_set_freq(pvr);

    ret |= pvr_set_standard(pvr);

    ret |= pvr_configure_audio(pvr);
    
    ret |= pvr_decoder_set_input(pvr);
    
    return ret;
}

/*
 * pvr_stop()
 *
 * Stop the stream. Call with lock held. After it, we can free all the buffer
 * without any danger, as long as we don't release the pvr->sem lock.
 */

static void pvr_stop(struct pvr *pvr)
{
    pvr_debug("pvr_stop");

    /* mask all interrupts */
    pvr_write_register(pvr, 0x0048, 0xffffffff);
    /* stop capture on encoder side */
    pvr_encoder_stop(pvr);
    /* stop capture */
    pvr_write_u8(pvr, 0x37, 0);
    /* change some GPIO data */
    pvr_write_register(pvr, 0x9020, 0x81040000);
    pvr_write_register(pvr, 0x900c, 0x08000000);
}

/*
 * pvr_start()
 *
 * Start the stream. Call with lock held.
 *
 */

PVR_DECLARE_CMD(start_capture, 0, "\x36\x03\x00\x00\x00\x00\x00\x44");

static void pvr_start(struct pvr *pvr)
{
    pvr_debug("pvr_start");

    /* unmask some interrupts */
    pvr_write_register(pvr, 0x0048, 0xbfffffff);
    /* change some GPIO data */
    pvr_write_register(pvr, 0x9020, 0x80040000);
    pvr_write_register(pvr, 0x900c, 0x08000000);

    /* Start captutre (in MPEG) */
    pvr_do_command(pvr, &start_capture);
    pvr_encoder_start(pvr);

    pvr->last_timestamp = jiffies;
}

/* State changing functions. Call with lock held. */

void pvr_state_change(struct pvr *pvr, enum pvr_device_state new_state)
{
    enum pvr_device_state old_state = pvr->dev_state;

    if ((old_state == _reset) && (new_state == _reset)){
        pvr->dev_state = _unstable;
        return;
    }

    if (old_state == new_state || old_state == _unstable)
        return;

    /* going to a lower state */

    if (new_state == _unstable)
        /* prevent any ctl io */
        pvr->dev_state = _unstable;

    if (new_state == _reset) {
        pvr_reset_ctl_endpoints(pvr);
        /* prevent recursive reset */
        pvr->dev_state = _reset;
    }

    if ((new_state <= _stopped) && (old_state > _stopped))
        pvr_stop(pvr);

    if (new_state <= old_state)
        goto done;

    /* changing to an higher state */

    switch (old_state){
        case _reset:
        case _need_firmware:
            if (new_state == _need_firmware)
                break;
            /* As it sometimes fails, give it 2 tries */
            if (pvr_upload_firmware2(pvr) && pvr_upload_firmware2(pvr)){
                err("device unstable!!");
                pvr->dev_state = _unstable;
                return;
            }
        case _need_config:
            if (new_state == _need_config)
                break;
            pvr_configure_capture(pvr);
        case _stopped:
            if (new_state == _stopped)
                break;
            pvr_start(pvr);
        default:
            break;
    }

done:
    pvr->dev_state = new_state;

}

inline void pvr_state_above(struct pvr *pvr, enum pvr_device_state state)
{
    if (pvr->dev_state < state)
        pvr_state_change(pvr, state);
}

inline void pvr_state_below(struct pvr *pvr, enum pvr_device_state state)
{
    if (pvr->dev_state > state) 
        pvr_state_change(pvr, state);
}

inline void pvr_reconfigure(struct pvr *pvr)
{
    pvr_state_below(pvr, _need_config);
}

static int pvr_v4l_open(struct inode *, struct file *);
static int pvr_v4l_release(struct inode *, struct file *);
int pvr_v4l_do_ioctl(struct inode *, struct file *,
				 unsigned int, void *);
static int pvr_v4l_ioctl(struct inode *, struct file *,
		       unsigned int, unsigned long);
static ssize_t pvr_v4l_read(struct file *, char *, size_t,
                            loff_t *);
static unsigned int pvr_v4l_poll(struct file *, poll_table *);
static int pvr_v4l_mmap(struct file *, struct vm_area_struct *);

static struct file_operations pvr_fops = {
	.owner		= THIS_MODULE,
	.open		= pvr_v4l_open,
	.release	= pvr_v4l_release,
	.read		= pvr_v4l_read,
	.ioctl		= pvr_v4l_ioctl,
	.llseek		= no_llseek,
    .poll       = pvr_v4l_poll,
    .mmap       = pvr_v4l_mmap,
};

static struct video_device pvr_video_template = {
	.owner		= THIS_MODULE,
	.type       = VID_TYPE_CAPTURE | VID_TYPE_TUNER,
    .type2      =  V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_TUNER | V4L2_CAP_AUDIO
                  | V4L2_CAP_READWRITE | V4L2_CAP_STREAMING,
    .hardware	= VID_HARDWARE_PVRUSB2,
	.fops		= &pvr_fops,
};

static void pvr_unregister_video(struct pvr *pvr)
{
	if (pvr->vdev) {
		info("unregister video /dev/video%d", pvr->vdev->minor & 0x1f);
		if (pvr->vdev->minor != -1)
			video_unregister_device(pvr->vdev);
		else
			video_device_release(pvr->vdev);
		pvr->vdev = NULL;
	}
}

/* register video4linux device */
static int __devinit pvr_register_video(struct pvr *pvr)
{
	struct video_device *vdev;

    if (pvr->udev == NULL) {
		err("%s: pvr->dev is not set", __FUNCTION__);
		goto exit;
	}

	vdev = video_device_alloc();
	if (vdev == NULL){
		err("%s: alloc of vdev failed", __FUNCTION__);;
        goto exit;
    }
	
	memcpy(vdev, &pvr_video_template, sizeof(struct video_device));
	pvr->vdev = vdev;
    vdev->dev = &pvr->udev->dev;
	vdev->release = video_device_release;
	video_set_drvdata(vdev, pvr);
    
	if (video_register_device(pvr->vdev, VFL_TYPE_GRABBER, video_nr))
		goto exit;

    info("pvrusb2: registered video device /dev/video%d",
				pvr->vdev->minor & 0x1f);
    
    return 0;

 exit:
	err("pvrusvb2: unable to register video device /dev/video%d", video_nr);
	pvr_unregister_video(pvr);
	return -1;
}


/*
 * pvr_open()
 *
 * This is part of Video 4 Linux API. The driver can be opened by multiple
 * clients (checks internal counter 'pvr->user'). This helps using v4lctl while
 * playing /dev/video?.
 * The procedure then allocates buffers needed for video processing.
 *
 */

static int pvr_v4l_open(struct inode *inode, struct file *file)
{
    struct video_device *vdev = video_devdata(file);
    struct pvr *pvr = (struct pvr *) video_get_drvdata(vdev);
    int ret = 0;

    pvr_debug("pvr_v4l_open");

    down(&pvr->sem);

    if (pvr->disconnecting){
        up(&pvr->sem);
        return -EIO;
    }

    if (pvr->user > 0){
        /* work has been done by previous users. */
        pvr->user++;
        up(&pvr->sem);
        return 0;
    }

    /* upload firmware encoder */
    pvr_state_above(pvr, _need_config);
/*
    ret = pvr_alloc_buffer(pvr);
    if (ret){
        err("alloc of video buffers failed");
        pvr_free_buffer(pvr);
        up(&pvr->sem);
        return ret;
    }
*/
    pvr->user++;

    up(&pvr->sem);

    return ret;
}

/*
 * pvr_release_final()
 *
 * This code does final release of struct pvr. This happens
 * after the device is disconnected and clients closed their files.
 *
 */
static void pvr_release_final(struct pvr *pvr)
{
	pvr_debug("pvr_final_release");

    if (pvr == NULL)
        return;

    pvr_unregister_video(pvr);

	kfree(pvr);
}

/*
 * pvr_v4l_release()
 *
 * This is part of Video 4 Linux API. The procedure
 * stops streaming and deallocates all buffers that were earlier
 * allocated in pvr_v4l_open().
 *
 */

static int pvr_v4l_release(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct pvr *pvr = (struct pvr *) video_get_drvdata(vdev);

    pvr_debug("pvr_v4l_release");

    down(&pvr->sem);

    pvr->user--;

    if(pvr->user > 0){
        /* work will be done  by someone else */
        up(&pvr->sem);
        return 0;
    }
    /* stop the stream... */
    pvr_state_below(pvr, _stopped);
    /* then free all the buffers */
    if (pvr->mem_state == _mapped)
        err("still mapped AARGH!!!");
    pvr_unlink_all_buffer(pvr);
    pvr_free_buffer(pvr);
    pvr->io_method = PVR_IO_READ; 
    up(&pvr->sem);

    if (pvr->disconnecting){
		pvr_debug("%s: Final disconnect", __FUNCTION__);
        pvr_release_final(pvr);
	}

    return 0;
}

static int pvr_v4l_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{

/* Temporary hack : use ivtv api until a v4l2 one is available. */
#define IVTV_IOC_G_CODEC        0xFFEE7703
#define IVTV_IOC_S_CODEC        0xFFEE7704
    if (cmd == IVTV_IOC_G_CODEC || cmd == IVTV_IOC_S_CODEC)
        return 0;
    return video_usercopy(inode, file, cmd, arg, pvr_v4l_do_ioctl);
}

static ssize_t pvr_v4l_read(struct file *file, char *buff, size_t count,
                            loff_t *ppos)
{
    struct video_device *vdev = video_devdata(file);
    struct pvr *pvr = (struct pvr *) video_get_drvdata(vdev);
    struct pvr_buffer *b;
    int i, n, retry = 1;

//    pvr_debug("pvr_v4l_read : %i", count);

    if (pvr->io_method != PVR_IO_READ)
        return -EBUSY;

    if (!count) 
        return 0;

    down(&pvr->sem);

    if (!pvr->user || pvr->disconnecting || !pvr->udev){
        up(&pvr->sem);
        return -EIO;
    }

    /* first read */
    if (pvr->dev_state < _streaming) {
        if (pvr->mem_state == _freed)
            pvr_alloc_buffer(pvr);
        pvr_submit_all_buffer(pvr);
        pvr_state_change(pvr, _streaming);
    }
re:
    /* take data in the rx_buffer ring */
    i = pvr_buffer_wait(pvr, (file->f_flags & O_NONBLOCK) ? 0 : 2 * HZ);

    if (i < 0) {
        /* nothing for us. it can be -EIO or -EAGAIN */
        if ((i == -EIO) && retry) {
            /* retry after a reset */
            retry = 0;
            pvr_state_below(pvr, _need_config);
            pvr_state_above(pvr, _streaming);
            goto re;
        }        
        up(&pvr->sem);
        return i;
    }

    b = pvr->rx_buffer[i];

    if (b == NULL){
        err("ring is fucked !!!!");
        up(&pvr->sem);
        return -EIO;
    }

    n = min(b->size, count);

    copy_to_user(buff, b->data, n);

    if(n == b->size) {
        /* we took all the data in this buffer. Re-submit it. */
        pvr->rx1++;
        pvr_submit_buffer(pvr, i);
    } else {
        /* Some data remains in this urb. Keep it. */
        b->data += n;
        b->size -= n;
    }

    up(&pvr->sem);

    return n;
}

static int pvr_v4l_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct video_device *dev = video_devdata(file);
	struct pvr *pvr = (struct pvr *) video_get_drvdata(dev);
    int ret;

    down(&pvr->sem);
    ret = pvr_map_buffer(pvr, vma);
    up(&pvr->sem);

	return ret;
}

/* FIXME : Not sure if this one is correct ! */

static unsigned int pvr_v4l_poll(struct file *file, poll_table *wait)
{
    struct video_device *vdev = video_devdata(file);
    struct pvr *pvr = (struct pvr *) video_get_drvdata(vdev);
	unsigned int mask = 0;

    down(&pvr->sem);
    pvr_state_above(pvr, _streaming);
    up(&pvr->sem);

	/* add stream's waitq to poll list */
	poll_wait(file, &pvr->wait_data, wait);

	mask |= POLLIN | POLLRDNORM;

	return mask;
}

/*
 * pvr_probe()
 *
 * This procedure queries device descriptor and accepts the interface
 * if it is a WINTV-PVR USB2 video device.
 *
 */

static void pvr_configure_pvr(struct pvr *pvr)
{

#define PVR_DEFAULT_REG(id) (pvr_regular_ctl[id-V4L2_CID_BASE].default_value)
#define PVR_DEFAULT_PRIV(id) (pvr_private_ctl[id-V4L2_CID_PRIVATE_BASE].default_value)

    pvr->standard = PVR_STD_SECAM;
    pvr->video_input = 0;
    pvr->audio_input = 0;
    pvr->video_width = PVR_WIDTH_DVD;
    pvr->video_height = (pvr->standard == PVR_STD_NTSC) ? 480 : 576;
    pvr->frequency = 7796;
    pvr->frequency = 12788; // 6
    // pvr->frequency = 3072; // 4
    pvr->video_bitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VIDEOBITRATE);
    pvr->video_vbr = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VBR);
    pvr->video_peakbitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VIDEOPEAK);
    pvr->video_brightness = PVR_DEFAULT_REG(V4L2_CID_BRIGHTNESS);
    pvr->video_contrast = PVR_DEFAULT_REG(V4L2_CID_CONTRAST);
    pvr->video_saturation = PVR_DEFAULT_REG(V4L2_CID_SATURATION);
    pvr->video_hue = PVR_DEFAULT_REG(V4L2_CID_HUE);
    pvr->audio_mode = V4L2_TUNER_MODE_STEREO;
    pvr->audio_bitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOBITRATE);
    pvr->audio_srate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_SRATE);
    pvr->audio_crc = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOCRC);
    pvr->audio_emphasis = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOEMPHASIS);
    pvr->audio_layer = 2;
    pvr-> audio_lang = 1;
    pvr->audio_muted = PVR_DEFAULT_REG(V4L2_CID_AUDIO_MUTE);
    pvr->audio_volume = PVR_DEFAULT_REG(V4L2_CID_AUDIO_VOLUME);
    pvr->audio_balance = PVR_DEFAULT_REG(V4L2_CID_AUDIO_BALANCE);
    pvr->audio_bass = PVR_DEFAULT_REG(V4L2_CID_AUDIO_BASS);
    pvr->audio_treble = PVR_DEFAULT_REG(V4L2_CID_AUDIO_TREBLE);
}

static int pvr_probe(struct usb_interface *intf, const struct usb_device_id *devid)
{
	struct usb_device *udev = interface_to_usbdev(intf);
	__u8 ifnum = intf->cur_altsetting->desc.bInterfaceNumber;
	struct pvr *pvr = NULL;
    int ret;

	info("VID=%#04x, PID=%#04x, ifnum=%u",udev->descriptor.idVendor,
         udev->descriptor.idProduct, ifnum);

    info("endpoints : %i",intf->cur_altsetting->desc.bNumEndpoints);

    usb_set_interface(udev, ifnum, 0);

 	pvr = kmalloc(sizeof(struct pvr), GFP_KERNEL);
    if (pvr == NULL) {
		err("%s: couldn't allocate pvr struct", __FUNCTION__);
		return -ENOMEM;
	}

    memset(pvr, 0, sizeof(struct pvr));

    /*
     * If the FX2 firmware is not loaded, do it. Then the device will
     * re-enumerate, closing this driver instance.
     */
    if (intf->cur_altsetting->desc.bNumEndpoints == 0) {
        pvr->renum = 1;
        usb_set_intfdata(intf, pvr);
        pvr_upload_firmware1(udev, FIRMWARE1_FILE);
        return 0;
    }

    pvr->udev = udev;
    init_MUTEX(&pvr->sem);
    init_MUTEX(&pvr->ctl_sem);

    init_waitqueue_head (&pvr->wait_data);
    init_waitqueue_head (&pvr->urbs_collected);
    spin_lock_init (&pvr->ring_lock);

    pvr->nr = pvr_nr++;
    pvr->dev_state = _need_firmware;
    pvr->user = 0;
    pvr->buffer_count = PVR_NBUFF;
    pvr->io_method = PVR_IO_READ;
    pvr->mem_state = _freed;

    pvr_configure_pvr(pvr);

    ret = pvr_register_video(pvr);
    if (ret){
        kfree(pvr);
        return ret;
    }

    ret = pvr_input_init(pvr);
    if (ret){
        kfree(pvr);
        return ret;
    }
    usb_set_intfdata(intf, pvr);
    intf->minor = pvr->vdev->minor;
    pvr_reset_ctl_endpoints(pvr);
    pvr_eeprom_analyse(pvr);
    pvr_input_start(pvr);

    return 0;
}

/*
 * pvr_disconnect()
 *
 */

static void pvr_disconnect(struct usb_interface *intf)
{
    struct pvr *pvr = usb_get_intfdata(intf);

    pvr_debug("pvr_disconnect");

    pvr->disconnecting = 1;

    if(pvr->renum){
        /* we did not initialize everything because device was 
         * going to re-enumerate
         */
        kfree(pvr);
        return;
    }

    down(&pvr->sem);

    /* Stop video io and remote stuff. */
    pvr_state_below(pvr, _stopped);
    pvr_input_stop(pvr);
    pvr_input_clear(pvr);

    usb_set_intfdata (intf, NULL);

    if (pvr->user) {
        pvr_debug("In use, disconnect pending");
        up(&pvr->sem);
    } else  {
        up(&pvr->sem); // needed ?
        pvr_release_final(pvr);
	}
}

static struct usb_driver pvr_driver = {
    owner:   THIS_MODULE,
    name:           "pvrusb2",
    id_table:       pvr_table,
    probe:          pvr_probe,
	disconnect:     pvr_disconnect
};

/*
 * pvr_init() / pvr_exit()
 *
 * This code is run to initialize/exit the driver.
 *
 */

static int __init pvr_init(void)
{
	int ret;

    pvr_debug("pvr_init");
 
    ret = usb_register(&pvr_driver);

	if (ret == 0)
		info(DRIVER_DESC " : " DRIVER_VERSION);

	return ret;
}

static void __exit pvr_exit(void)
{
    
    pvr_debug("pvr_exit");

    usb_deregister(&pvr_driver);
}

module_init(pvr_init);
module_exit(pvr_exit);

MODULE_DEVICE_TABLE (usb, pvr_table);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
