/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __PVRUSB2_H
#define __PVRUSB2_H


#include <linux/videodev.h>
#include <linux/usb.h>

/* size of a firmware chunk */
#define FIRMWARE_CHUNK_SIZE 0x2000

/* Various files we will load with firmware_entry */
#define FIRMWARE1_FILE  "pvrusb2.f1"
#define FIRMWARE2_FILE  "pvrusb2.f2"

/* Buffer management constants */
#define PVR_VID_BUFFSIZE        PAGE_ALIGN(0x8000)
#define PVR_CTL_BUFFSIZE        0x0100
#define PVR_NBUFF               16       /* Max number of video buffers */

#define PVR_VID_ENDPOINT        0x84
#define PVR_CTL_WRITE_ENDPOINT  0x01
#define PVR_CTL_READ_ENDPOINT   0x81
#define PVR_FIRMWARE_ENDPOINT   0x02

#define PVR_DEBUG 1
#define pvr_debug(...)    do { if(PVR_DEBUG) info(__VA_ARGS__); } while(0)
/* BUG_ON is a bit aggressive... */
#define PVR_BUG_ON(cond)   do { if (cond) err("BUG !"); } while(0)

#define VID_HARDWARE_PVRUSB2    38  /* FIXME : need a good value */

/* IO method supported */
#define PVR_IO_READ 0
#define PVR_IO_MMAP 1


#define PVR_DECOMPOSE_LE(t,i,d)   (t)[i] = (d) & 0xff;\
                                  (t)[i+1] = ((d) >> 8) & 0xff;\
                                  (t)[i+2] = ((d) >> 16) & 0xff;\
                                  (t)[i+3] = ((d) >> 24) & 0xff

#define PVR_DECOMPOSE_BE(t,i,d)   (t)[i+3] = (d) & 0xff;\
                                  (t)[i+2] = ((d) >> 8) & 0xff;\
                                  (t)[i+1] = ((d) >> 16) & 0xff;\
                                  (t)[i] = ((d) >> 24) & 0xff

enum pvr_device_state {
    _unstable=0,        /* device is fucked up. Absorbant state. */
    _reset,
    _need_firmware,     /* state 0 */
    _need_config,       /* encoder firmware in. Need to configure capture. */
    _stopped,           /* Device is ready to start capture. */
//    _should_stream,
//    _should_stream_m,     /* we should deliver an MPEG stream */
//    _should_stream_mv,   /* we should deliver MPEG and VBI streams */
//    _streaming_m,         /* we deliver an MPEG stream */
//    _streaming_v,       /* we deliver VBI stream */
//    _streaming_mv,       /* we deliver MPEG+VBI streams */
    _streaming
};

enum pvr_memory_state {
    _freed,
    _allocated,
    _mapped,    /* at least one buffer is mapped */
};

struct pvr_cmd {
    u8 length; 
    u8 ret;
    u8 *cmd;
};

#define PVR_SIZEOf_CMD(c)   (2*sizeof(short)+sizeof(char)*(c)->length)
#define PVR_CMD(r,c)   { .length = sizeof(c)-1, .ret = r , .cmd = c }
#define PVR_DECLARE_CMD(name,r,c) struct pvr_cmd name = PVR_CMD(r,c)

struct pvr_mapping {
	struct pvr *pvr;
	int count;
};

struct pvr_buffer {
    struct pvr *pvr;
    void *buff;     /* Allocated buffer. Page aligned one is in purb */
    void *data;     /* First unread data byte */
    size_t size;    /* Size of unread data (for read-io) */
    struct urb *purb;
    short submitted;
    struct pvr_mapping *map;    /* NULL if not mapped */
};

struct pvr {
    struct video_device *vdev;
    struct usb_device* udev;
    struct semaphore sem;
    int user;
    int nr;
    int disconnecting;
    int renum;
    unsigned long last_timestamp;   /* last buffer timestamp */
    enum pvr_device_state dev_state;
    enum pvr_memory_state mem_state;
//    int reset;  /* to detect when a reset failed */
    int io_method;  /* change under lock */
    int tuner;
    int supported_std;

    atomic_t submitted_urbs; /* when trying to stop device */
    wait_queue_head_t urbs_collected;
    wait_queue_head_t wait_data;

    struct pvr_buffer buffer[PVR_NBUFF];        /* allocated buffers */
    int buffer_count;   /* how many buffer do we want ? */
    int map_count;      /* number of mapped buffers */
    struct pvr_buffer *rx_buffer[PVR_NBUFF];    /* ring of recv buffers */
    spinlock_t ring_lock;   /* we don't want simultaneous add to the ring. */
    unsigned int rx1;
    unsigned int rx2;

    /* our control pool */
    struct semaphore ctl_sem;
    unsigned char cmd_buffer[PVR_CTL_BUFFSIZE];  /* used under ctl_sem lock */

    /* remote stuff */
    struct input_dev* idev;
    struct timer_list remote_timer;    /* will read key pressed on the remote. */
    struct work_struct remote_task;
    struct semaphore task_sem;
    int last_key;

    /* audio/video parameters */
    unsigned int frequency;
    int video_input;
    int audio_input;
    int standard;
    int video_width;
    int video_height;
    int video_bitrate;
    int video_peakbitrate;
    int video_vbr;
    int video_brightness;
    int video_contrast;
    int video_saturation;
    int video_hue;

    int audio_status_bilingual;
    int audio_status_stereo;
    int audio_lang;
    int audio_mode;
    int audio_bitrate;
    int audio_srate;
    int audio_layer;    /* only layer 2 seems supported */
    int audio_crc;
    int audio_emphasis;
    int audio_muted;
    int audio_volume;
    int audio_balance;
    int audio_bass;
    int audio_treble;
};

int pvr_send_sequence(struct pvr *, struct pvr_cmd *);
int pvr_send_request(struct pvr *, void *, unsigned int, void *, unsigned int);
int pvr_do_command(struct pvr *, struct pvr_cmd *);

int pvr_write_register(struct pvr *, u16, u32);
int pvr_read_register(struct pvr *, u16, u32 *);
int pvr_write_u16(struct pvr *, u16, int);
int pvr_write_u8(struct pvr *, u8, int);
int pvr_i2c_write(struct pvr *, u8, u8 *, u8);
int pvr_i2c_read(struct pvr *, u8, u8 *, u8, u8 *, u8);

inline void pvr_reconfigure(struct pvr *);
void pvr_state_change(struct pvr *, enum pvr_device_state);
inline void pvr_state_above(struct pvr *, enum pvr_device_state);
inline void pvr_state_below(struct pvr *, enum pvr_device_state);

#endif /* __PVRUSB2_H */
